import json
import logging
import math
import re
import traceback
from typing import Final, List, Tuple
from urllib.parse import unquote_plus

import requests
from bs4 import BeautifulSoup, Tag
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from .artwork import ArtworkMetadata, CategoryType, ImageMetadata

logger = logging.getLogger(__name__)


class ArtveeClient:
    """HTTP client for interacting with the Artvee API.

    Constants:
        _HTTP_CONN_TIMEOUT_SEC (float):
            Default number of seconds to wait to establish a connection to a remote machine.
        _HTTP_READ_TIMEOUT_SEC (float):
            Default number of seconds the client will wait for the server to send a response.
        _ITEMS_PER_PAGE (int):
            Maximum number of items to retrieve per page in API requests.
        _TITLE_PATTERN (re.Pattern):
            Regex pattern for extracting title and date. ex: `Landscape with Weather Vane (1935)`;
            group 1 = title (ex: Landscape with Weather Vane), group 2 = date (ex: 1935)
        _ARTIST_PATTERN (re.Pattern):
            Regex pattern for extracting artist name and origin. ex: `Arthur Dove(American, 1880-1946)`;
            group 1 = artist name (ex: Arthur Dove), group 2 = origin (ex: American, 1880-1946)
        _RESOURCE_PATTERN (re.Pattern):
            Regex pattern for extracting the resource name. ex: `https://artvee.com/dl/zwei-tanzende/`;
            group 1 = resource (ex: zwei-tanzende)
        _IMG_DIMENSION_PATTERN (re.Pattern):
            Regex pattern for extracting image dimensions. ex: `1800 x 1185px`;
            group 1 = width (ex: 1800), group 2 = height (ex: 1185)
        _IMG_FILE_SIZE_PATTERN (re.Pattern):
            Regex pattern for extracting image file size and unit.ex: `1.82 MB`; group 1 = size (ex: 1.82), group 2 = unit (ex: MB)

    Attributes:
        _timeout (tuple[float, float]):
            Timeouts to use for HTTP requests.
        _session (Session):
            Allows persistance of parameters across HTTP requests.

    Args:
        conn_timeout (float, optional):
            Number of seconds to wait to establish a connection to a remote machine. Defaults 3.05 seconds.
        read_timeout (float, optional):
            Number of seconds the client will wait for the server to send a response. Defaults to 10 seconds.
        max_attempts (int, optional):
            The maximum number of attempts (including the initial call). Must be between 1 and 10.
            Defaults to 3 (initial call + two retries).

    Raises:
        ValueError:
            If `conn_timeout` is not positive.
            If `read_timeout` is not positive.
            If `max_attempts` is not in the range [1, 10].
    """

    _HTTP_CONN_TIMEOUT_SEC: Final[float] = 3.05
    _HTTP_READ_TIMEOUT_SEC: Final[float] = 10

    _ITEMS_PER_PAGE: Final[int] = 70

    _TITLE_PATTERN: Final[re.Pattern] = re.compile("^(.+) *\\((.+)\\) *$")
    _ARTIST_PATTERN: Final[re.Pattern] = re.compile("^(.+) *\\((.+)\\) *$")
    _RESOURCE_PATTERN: Final[re.Pattern] = re.compile(
        "^https://artvee\\.com/dl/((\\w|-|%)+)/$"
    )
    _IMG_DIMENSION_PATTERN: Final[re.Pattern] = re.compile("^(\\d+)\\sx\\s(\\d+)px$")
    _IMG_FILE_SIZE_PATTERN: Final[re.Pattern] = re.compile(
        "^((?:[0-9]*\\.)?[0-9]+)\\s([A-Za-z]+)$"
    )

    def __init__(
        self,
        conn_timeout: float = _HTTP_CONN_TIMEOUT_SEC,
        read_timeout: float = _HTTP_READ_TIMEOUT_SEC,
        max_attempts: int = 3,
    ) -> None:
        """Initializes a newly created `ArtveeClient` object."""

        if conn_timeout < 0:
            raise ValueError("Connection timeout cannot be a negative number")
        if read_timeout < 0:
            raise ValueError("Read timeout cannot be a negative number")
        self._timeout = (conn_timeout, read_timeout)

        if not 1 <= max_attempts <= 10:
            raise ValueError("Max attempts must be in range [1-10]")

        retry_config = Retry(
            total=max_attempts - 1,  # number of retry attempts
            backoff_factor=1,  # seconds
            backoff_jitter=True,
            status_forcelist=[502, 503, 504],  # transient failures
        )
        self._session = ArtveeClient._new_session(retry_config)

    def get_page_count(self, category: CategoryType) -> int:
        """Retrieve the total number of webpages for a given category.

        Args:
            category (CategoryType):
                The category for which to retrieve the page count.

        Returns:
            int: The total number of pages available for the specified category.

        Raises:
            requests.exceptions.HTTPError:
                If the HTTP request returns an unsuccessful status code.
            ValueError:
                If the total items cannot be parsed / converted to an integer.
        """

        logger.debug("Retrieving page count; category=%s", category)
        url = f"https://artvee.com/c/{category}/page/1/?per_page={self._ITEMS_PER_PAGE}"

        resp = self._session.get(
            url,
            timeout=self._timeout,
        )

        resp.raise_for_status()

        soup = BeautifulSoup(resp.content, "html.parser")

        total_items = (
            soup.find("p", class_="woocommerce-result-count")
            .text.strip("items")
            .strip()
        )

        return math.ceil(int(total_items) / self._ITEMS_PER_PAGE)

    def get_metadata(
        self, category: CategoryType, page: int
    ) -> List[Tuple[ArtworkMetadata, ImageMetadata]]:
        """Retrieve artwork metadata for a specified category and page.

        Args:
            category (CategoryType):
                The category for which to retrieve artwork metadata.
            page (int):
                The page number to retrieve the metadata from. Pages are indexed starting at 1.

        Returns:
            List[Tuple[ArtworkMetadata, ImageMetadata]]:
                A list where each tuple represents the attributes of an artwork.
                `ArtworkMetadata`: Attributes of an artwork.
                `ImageMetadata`: Attributes of an image file.

        Raises:
            requests.exceptions.HTTPError:
                If the HTTP request returns an unsuccessful status code.
        """

        logger.debug("Retrieving metadata; category=%s, page=%d", category, page)
        page_url = f"https://www.artvee.com/c/{category}/page/{page}/?orderby=title_asc&per_page={self._ITEMS_PER_PAGE}"

        resp = self._session.get(
            page_url,
            timeout=self._timeout,
        )

        resp.raise_for_status()

        soup = BeautifulSoup(resp.content.decode("utf-8"), "html.parser")
        all_metadata_html = soup.find_all("div", {"class": "product-element-bottom"})

        metadata_tuples = []
        for metadata_html in all_metadata_html:
            try:
                artwork_metadata = ArtveeClient._parse_artist_metadata(
                    metadata_html, category
                )
                image_metadata = ArtveeClient._parse_image_metadata(metadata_html)

                metadata_tuples.append((artwork_metadata, image_metadata))
            except Exception:
                logger.warning(traceback.format_exc())
                pass

        return metadata_tuples

    def get_image(self, img_metadata: ImageMetadata) -> bytes:
        """Retrieve the image data.

        Args:
            img_metadata (ImageMetadata):
                Information that describes attributes of an artwork.

        Returns:
            bytes:
                The raw JPG image data.

        Raises:
            requests.exceptions.HTTPError:
                If the HTTP request returns an unsuccessful status code.
        """

        logger.debug("Retrieving image; url=%s", img_metadata.source_url)
        get_img_resp = self._session.get(
            img_metadata.source_url,
            timeout=self._timeout,
        )

        get_img_resp.raise_for_status()
        return get_img_resp.content

    @staticmethod
    def _new_session(retry_config: Retry) -> requests.Session:
        session = requests.Session()

        adapter = HTTPAdapter(max_retries=retry_config)
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        return session

    @staticmethod
    def _parse_artist_metadata(
        metadata_html: Tag, category: CategoryType
    ) -> ArtworkMetadata:
        artist_details = metadata_html.find("h3", {"class": "product-title"})
        title = artist_details.get_text(strip=True)
        url = artist_details.a.get("href")
        resource = ArtveeClient._RESOURCE_PATTERN.match(url).group(1)
        resource = unquote_plus(resource)

        artwork_metadata = ArtworkMetadata(
            url, resource, title, category.value.capitalize()
        )

        if title_matcher := ArtveeClient._TITLE_PATTERN.match(title):
            artwork_metadata.title = title_matcher.group(1).strip()
            artwork_metadata.date = title_matcher.group(2).strip()

        artwork_metadata.artist = metadata_html.find(
            "div", {"class": "woodmart-product-brands-links"}
        ).get_text(strip=True)

        if artist_matcher := ArtveeClient._ARTIST_PATTERN.match(
            artwork_metadata.artist
        ):
            artwork_metadata.artist = artist_matcher.group(1).strip()
            artwork_metadata.origin = artist_matcher.group(2).strip()

        return artwork_metadata

    @staticmethod
    def _parse_image_metadata(metadata_html: Tag) -> ImageMetadata:
        img_details_json = metadata_html.find("div", {"class": "tbmc linko"}).get(
            "data-sk"
        )
        img_details = json.loads(img_details_json)

        img_metadata = ImageMetadata()

        sdl_image_size = img_details.get("sdlimagesize")
        if sdl_dimension_matcher := ArtveeClient._IMG_DIMENSION_PATTERN.match(
            sdl_image_size
        ):
            img_metadata.width = int(sdl_dimension_matcher.group(1))
            img_metadata.height = int(sdl_dimension_matcher.group(2))

        sdl_file_size = img_details.get("sdlfilesize")
        if sdl_file_size_matcher := ArtveeClient._IMG_FILE_SIZE_PATTERN.match(
            sdl_file_size
        ):
            img_metadata.file_size = float(sdl_file_size_matcher.group(1))
            img_metadata.file_size_unit = sdl_file_size_matcher.group(2)

        if sk := img_details.get("sk"):
            img_metadata.source_url = f"https://mdl.artvee.com/sdl/{sk}sdl.jpg"

        return img_metadata



from __future__ import annotations

import base64
from dataclasses import dataclass
from enum import Enum


@dataclass
class ArtworkMetadata:
    """Information that describes attributes of an artwork.

    Attributes:
        url (str):
            Artwork URL (ex: https://artvee.com/dl/zwei-tanzende/)
        resource (str):
            Unique name of artwork; extracted from the URL (ex: zwei-tanzende)
        title (str):
            Name of the artwork (ex: Zwei Tanzende)
        category (str):
            Category the work of art is depicting (ex: abstract)
        artist (str, optional):
            Name of the person that created this artwork. Defaults to `Unknown Artist`
        date (str, optional):
            Year or date range the artwork was completed (ex: 2012 - 2019). Defaults to `n.d.`, no date
        origin (str, optional):
            Artist nationality and lifespan (ex: Austrian, 1834-1921). Defaults to `None`
    """

    url: str
    resource: str
    title: str
    category: str
    artist: str = "Unknown Artist"
    date: str = "n.d."
    origin: str | None = None


@dataclass
class Artwork(ArtworkMetadata):
    """Represents an artistic work.

    Attributes:
        image (Image, optional):
            The image, including associated metadata. Defaults to `None`
    """

    image: Image | None = None

    def to_dict(self) -> dict | None:
        """A dict representation of this object.

        The image will be encoded in Base64 binary.

        Returns:
            This object as a dict.
        """
        return dict(self.__dict__, image=self.image.to_dict() if self.image else None)


@dataclass
class ImageMetadata:
    """Information that describes attributes of an image file.

    Attributes:
        source_url (str, optional):
            The URL where the image is sourced from. Defaults to `None`
        width (int, optional):
            The width of the image in pixels. Defaults to `0`
        height (int, optional):
            The height of the image in pixels. Defaults to `0`
        file_size (float, optional):
            The size of the image file. Defaults to `0`
        file_size_unit (str, optional):
            The unit of the file size (ex: "KB", "MB"). Defaults to `None`
    """

    source_url: str | None = None
    width: int = 0
    height: int = 0
    file_size: float = 0
    file_size_unit: str | None = None


@dataclass
class Image(ImageMetadata):
    """Represents a graphical image.

    Attributes:
        raw (bytes, optional):
            The raw binary data of the image. Defaults to `None`
        format_name (str, optional):
            The format of the image file - identifies how the image should be processed/displayed. Defaults to `jpg`.
    """

    raw: bytes | None = None
    format_name: str = "jpg"

    def b64encoded(self) -> str | None:
        """Returns the image encoded as Base64 binary.

        Returns:
            Base64 binary encoded image.
        """
        return (
            str(base64.b64encode(self.raw).decode("utf-8"))
            if self.raw is not None
            else None
        )

    def to_dict(self) -> dict | None:
        """A dict representation of this object.

        The image will be encoded in Base64 binary.

        Returns:
            This object as a dict.
        """
        return dict(self.__dict__, raw=self.b64encoded())


class CategoryType(Enum):
    """Enumeration for different categories of art.

    Attributes:
        ABSTRACT (str):
            Art that uses shapes, colors, forms, and gestural marks rather than aiming for an accurate representation
            of visual reality.
        FIGURATIVE (str):
            Art that represents recognizable subjects, particularly the human figure, focusing on real-world forms.
        LANDSCAPE (str):
            Art that depicts natural scenes, often focusing on the beauty and atmosphere of the environment.
        RELIGION (str):
            Art that conveys spiritual themes or depicts subjects related to faith, spirituality, and religious practices.
        MYTHOLOGY (str):
            Art that illustrates or interprets themes, characters, and stories from myths and legends, often exploring
            cultural beliefs and narratives.
        POSTERS (str):
            Art designed for display and promotion, often featuring bold imagery and text to convey a message or advertise
            events, products, or causes.
        ANIMALS (str):
            Art that focuses on the representation of animals, capturing their form, behavior, and characteristics.
        ILLUSTRATION (str):
            Art that creates images to accompany text or convey a narrative, often found in books, magazines, and advertising.
        STILL_LIFE (str):
            Art that depicts inanimate objects, such as fruits, flowers, and everyday items.
        BOTANICAL (str):
            Art that focuses on the representation of plants and flowers, often emphasizing accuracy and detail to depict
            their beauty and scientific characteristics.
        DRAWINGS (str):
            Art form created using various mediums, such as pencil, charcoal, or ink, to render images through lines and
            shading, often capturing ideas, sketches, or detailed representations.
        ASIAN_ART (str):
            Art that represents styles and techniques from Asian cultures, often reflecting traditional themes, motifs,
            and cultural significance.
    """

    ABSTRACT = "abstract"
    FIGURATIVE = "figurative"
    LANDSCAPE = "landscape"
    RELIGION = "religion"
    MYTHOLOGY = "mythology"
    POSTERS = "posters"
    ANIMALS = "animals"
    ILLUSTRATION = "illustration"
    STILL_LIFE = "still-life"
    BOTANICAL = "botanical"
    DRAWINGS = "drawings"
    ASIAN_ART = "asian-art"

    def __str__(self):
        return self.value

    def __lt__(self, other):
        return self.value < other.value



import concurrent.futures
import logging
import traceback
from threading import Event, Lock, Thread
from typing import Callable, Tuple

from typing_extensions import Self

from .artvee_client import ArtveeClient
from .artwork import Artwork, ArtworkMetadata, CategoryType, Image, ImageMetadata

logger = logging.getLogger(__name__)


class ArtveeScraper:
    """A web scraper which concurrently extracts artwork from Artvee. Callbacks are notified asynchronously for
    each scraped artwork so that user-defined actions may be taken.

    Attributes:
        _artvee_client (ArtveeClient):
            An HTTP client for accessing artwork.
        _worker_pool (ThreadPoolExecutor):
            A pool of threads to manage concurrent scraping tasks.
        _categories (Tuple[CategoryType]):
            Category types to scrape.
        _boss_thread (Thread):
            The main thread responsible for executing the scraping logic; delegates tasks to workers.
        _stop_event (Event):
            Signal to indicate the scraping process should be halted.
        _listener_lock (Lock):
            A lock which provides access the listeners in a thread-safe manner.
        _listeners (dict):
            A hashset of callbacks to invoke asynchronously.

    Args:
        artvee_client (ArtveeClient, optional):
            An HTTP client for accessing artwork. Defaults to a new instance.
        worker_threads (int, optional):
            The number of worker threads to use for processing. Must be between 1 and 10. Defaults to 3.
        categories (Tuple[CategoryType], optional):
            Category types to scrape. Defaults to all categories.

    Raises:
        ValueError:
            If `worker_threads` is not in the range [1, 10].
    """

    def __init__(
        self,
        artvee_client: ArtveeClient = ArtveeClient(),
        worker_threads: int = 3,
        categories: Tuple[CategoryType] = tuple(CategoryType),
    ) -> None:
        """Initializes a newly created `ArtveeScraper` object."""

        self._artvee_client = artvee_client
        if not 1 <= worker_threads <= 10:
            raise ValueError("Worker threads must be in range [1-10]")
        self._worker_pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=worker_threads
        )
        self._categories = categories

        self._boss_thread = Thread(target=self._exec, args=())
        self._stop_event = Event()
        self._listener_lock = Lock()
        self._listeners = {}

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.join()

    def register_listener(
        self,
        on_event_listener: Callable[[Artwork, Exception | None], None],
    ) -> Self:
        """Registers a callback to be notified of events asynchronously.

        A callback may only be registered once; additional attempts to register the same callback will
        overwrite the previous registration. The `success` notification contains a fully populated `Artwork`,
        whereas the `failure` notification contains a partially populated `Artwork` and associated exception.

        Args:
            on_event_listener (Callable[[Artwork, Exception | None], None]):
                A callback function that will be notified async when an event occurs.

        Raises:
            ValueError:
                If the provided `on_event_listener` argument is not callable.

        Returns:
            `self` for method chaining
        """

        if not callable(on_event_listener):
            raise ValueError("On event listener callback must be a callable type")

        with self._listener_lock:
            tmp = self._listeners.copy()
            tmp[on_event_listener] = None  # dict as a hashset
            self._listeners = tmp
            return self

    def deregister_listener(
        self,
        on_event_listener: Callable[[Artwork, Exception | None], None],
    ) -> Self:
        """Deregisters a callback so that it will no longer be notified of events.

        Args:
            on_event_listener (Callable[[Artwork, Exception | None]], None]):
                The callback to no longer notify when an event occurs.

        Returns:
            `self` for method chaining
        """

        with self._listener_lock:
            tmp = self._listeners.copy()
            tmp.pop(on_event_listener, None)
            self._listeners = tmp
            return self

    def start(self) -> None:
        """Begin scraping artwork

        Returns:
            None
        """

        logger.debug("Starting")
        self._boss_thread.start()

    def join(self) -> None:
        """Blocks the calling thread until all active tasks have been completed.

        Returns:
            None
        """

        logger.debug("Waiting for all active tasks to complete")
        self._boss_thread.join()
        self._worker_pool.shutdown(wait=True, cancel_futures=False)

    def shutdown(self) -> None:
        """Initiates a shutdown in which running tasks are completed, but all pending tasks are canceled.

        Returns:
            None
        """

        logger.debug("Shutting down")
        self._stop_event.set()
        self._boss_thread.join()
        self._worker_pool.shutdown(wait=True, cancel_futures=True)

    def _exec(self) -> None:
        logger.debug("Executing scraper for categories %s", self._categories)

        for category in self._categories:
            if self._stop_event.is_set():
                return

            try:
                page_count = self._artvee_client.get_page_count(category)
                logger.debug("Category %s has %d page(s)", category, page_count)

                for page in range(1, page_count + 1):
                    if self._stop_event.is_set():
                        return

                    logger.debug(
                        "Processing category %s, page (%d/%d)",
                        category,
                        page,
                        page_count,
                    )

                    try:
                        # Scrape the page's metadata
                        metadata_tuples = self._artvee_client.get_metadata(
                            category, page
                        )

                        # Submit tasks to download the images; invoke the listeners for each
                        results = self._worker_pool.map(
                            self._worker_task, metadata_tuples
                        )

                        # Block until all submitted tasks for a page have completed
                        for _ in results:
                            pass
                    except Exception:
                        logging.warning(
                            "An error occured while processing page %d of category %s; %s",
                            page,
                            category,
                            traceback.format_exc(),
                        )
            except Exception:
                logging.error(
                    "An error occured while processing category %s; %s",
                    category,
                    traceback.format_exc(),
                )

    def _worker_task(
        self, metadata_tuple: Tuple[ArtworkMetadata, ImageMetadata]
    ) -> None:
        """Artwork processing task.

        Retrieves the image, coalescing it with the metadata into an `Artwork` object.
        If the image bytes are successfully retrieved, any registered listeners will be invoked with a fully populated `Artwork`.
        If an exception is encountered while retrieving the image, any registered listeners will be invoked with a partially
        populated `Artwork` along with the exception that was thrown.

        Args:
            metadata_tuple (Tuple[ArtworkMetadata, ImageMetadata]):
                A pair of related metadata which provides the attributes of an artwork.
                - `ArtworkMetadata`: Attributes of an artwork.
                - `ImageMetadata`: Attributes of an image file.

        Returns:
            None
        """

        artwork_metadata, img_metadata = metadata_tuple

        # Unpack the metadata to partially populate the artwork
        artwork = Artwork(**vars(artwork_metadata))
        artwork.image = Image(**vars(img_metadata))

        try:
            artwork.image.raw = self._artvee_client.get_image(img_metadata)

            self._invoke_listeners(artwork)  # artwork is fully populated
        except Exception as thrown:
            self._invoke_listeners(
                artwork, thrown
            )  # artwork is partially populated (missing raw image)

    def _invoke_listeners(
        self, artwork: Artwork, thrown: Exception | None = None
    ) -> None:
        """Invokes registered callback functions.

        Args:
            artwork (Artwork): The artwork. Partially populated if an exception has occurred.
            thrown (Exception, optional): The processing exception. Defaults to None.

        Returns:
            None
        """
        for listener in self._listeners.keys():
            listener(artwork, thrown)



from setuptools import find_packages, setup

requires = ["beautifulsoup4==4.12.3", "requests==2.32.3"]

setup(
    name="artvee-scraper",
    author="Zach Duclos",
    author_email="zduclos.github@gmail.com",
    description="Fetch public domain artwork from Artvee (https://www.artvee.com)",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/zduclos/artvee-scraper",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    package_data={"artvee_scraper": ["py.typed"]},
    install_requires=requires,
    license="MIT",
    scripts=[],
    zip_safe=True,
    python_requires=">= 3.10",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Operating System :: OS Independent",
    ],
    keywords="artvee, artwork, webscraper",
    project_urls={
        "Bug Reports": "https://github.com/zduclos/artvee-scraper/issues",
        "Source": "https://github.com/zduclos/artvee-scraper",
    },
)
